# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Widgets for the Application plugin."""

from .about import AboutDialog  # noqa
from .status import InAppAppealStatus  # noqa
